package dev.fitko.fitconnect.client;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.steps.sender.StartStep;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.net.URI;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class SenderClientTest {

    private Sender senderMock;
    private StartStep underTest;

    private static final LogCaptor logs = new LogCaptor();

    @BeforeEach
    public void setup() {
        this.senderMock = mock(Sender.class);
        this.underTest = SenderClient.build(this.senderMock);
        logs.clearEvents();
    }

    @Test
    void testSendSubmissionWithJsonData() throws Exception {

        // Given
        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema jsonSchema = new SubmissionSchema();
        jsonSchema.setSchemaUri(URI.create("https://json.mimetype"));
        jsonSchema.setMimeType(MimeType.APPLICATION_JSON);
        final Destination destination = getDestination(destinationId, jsonSchema);

        when(this.senderMock.getDestination(any())).thenReturn(destination);
        when(this.senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(this.senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(this.senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(this.senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());


        // When
        final var submission = this.underTest
                .newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testSendSubmissionWithXmlData() throws Exception {

        // Given
        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema xmlSchema = new SubmissionSchema();
        xmlSchema.setSchemaUri(URI.create("https://xml.mimetype"));
        xmlSchema.setMimeType(MimeType.APPLICATION_XML);
        final Destination destination = getDestination(destinationId, xmlSchema);

        when(this.senderMock.getDestination(any())).thenReturn(destination);
        when(this.senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(this.senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(this.senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(this.senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());


        // When
        final var submission = this.underTest.newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withXmlData("<xml><test>data</test></xml>")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testMimeTypeDetection() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();
        final File file1 = new File("src/test/resources/text_attachment_with_no_filetype");
        final File file2 = new File("src/test/resources/pdf_attachment_with_no_extension");

        // When
        final var submission = this.underTest.newSubmission()
                .withAttachments(List.of(file1, file2))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("Detected attachment mime-type text/plain");
        logs.assertContains("Detected attachment mime-type application/pdf");
    }


    @Test
    void testWithMissingData() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData(null)
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Data is mandatory, but was null");
    }

    @Test
    void testWithMissingDestinationId() throws Exception {

        // Given
        setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(null)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("DestinationId is mandatory, but was null");
    }

    @Test
    void testMissingDataMimeTypeSchema() throws Exception {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(this.senderMock.getDestination(any())).thenReturn(destination);
        when(this.senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(this.senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Required schema to send valid submission not found");
    }

    @Test
    void testWithMissingServiceTypeName() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType(null, "leika:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("ServiceType name was null");
    }

    @Test
    void testWithMissingServiceDescription() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Test", null, "leika:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("ServiceType description was null");
    }

    @Test
    void testWithMissingServiceTypeIdentifier() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Name", null)
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("LeikaKey was null");
    }

    @Test
    void testWithMissingServiceTypeDescription() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Name", null, "leika:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
    }

    @Test
    void testWithEmptyAttachments() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withAttachments(List.of())
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentFileNull() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withAttachment(null)
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentPathNotPresent() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = this.underTest.newSubmission()
                .withAttachment(new File("/non/existing/path"))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Reading file failed. Attachment will not be created.");
    }

    @Test
    void testEncryptionError() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(this.senderMock.encryptBytes(any(), any())).thenThrow(new EncryptionException("Encryption failed"));

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Encrypting submission failed");
    }

    @Test
    void testAnnouncingSubmissionFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(this.senderMock.createSubmission(any())).thenThrow(new SubmissionNotCreatedException("Announcing submission failed"));

        // When
        final var submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Failed to announce new submission");

    }

    @Test
    void testRestApiCallFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(this.senderMock.getDestination(any())).thenThrow(new RestApiException("Loading destination failed"));

        // When
        final Optional<SentSubmission> submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Sending submission failed");

    }

    @Test
    void testGettingEncryptionKeyFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenThrow(new KeyNotRetrievedException("Getting encryption key failed"));

        // When
        final Optional<SentSubmission> submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Getting encryption key for destination " + destinationId + " failed");
    }

    @Test
    void testSendSubmissionWithEncryptedData() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final Optional<SentSubmission> submission = this.underTest.newSubmissionWithEncryptedData()
                .withEncryptedAttachments(Map.of(UUID.randomUUID(), "@tt@chm$nt"))
                .withEncryptedJsonData("$ncrypt$d d@t@")
                .withEncryptedMetadata("$ncrypt$d m$t@d@t@")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isPresent());
        logs.assertContains("SUCCESSFULLY HANDED IN SUBMISSION");
    }

    @Test
    void testPublicEncryptionKeyRetrieval() throws Exception {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(this.senderMock.getDestination(any())).thenReturn(destination);
        when(this.senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final Optional<String> encryptionKey = this.underTest.getPublicKey(destinationId);

        assertTrue(encryptionKey.isPresent());
        assertThat(encryptionKey.get(), equalTo(publicKey.toJSONString()));
    }

    @Test
    void testFailOnInvalidMetadata() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(this.senderMock.validateMetadata(any())).thenReturn(ValidationResult.error(new ValidationException("invalid metadata")));

        // When
        final Optional<SentSubmission> submission = this.underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Metadata does not match schema");
    }

    @Test
    void testGetValidEventLog() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        final List<EventLogEntry> expectedLog = List.of(EventLogEntry.builder().build());
        when(this.senderMock.getEventLog(any(), any())).thenReturn(expectedLog);

        // When
        final List<EventLogEntry> eventLog = this.underTest.getEventLog(caseId, destinationId);

        // Then
        assertThat(eventLog, is(expectedLog));

    }

    @Test
    void testGetInvalidEventLog() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        when(this.senderMock.getEventLog(any(), any())).thenThrow(new EventLogException("Invalid log"));

        // When
        final EventLogException exception = assertThrows(EventLogException.class, () -> this.underTest.getEventLog(caseId, destinationId));

        // Then
        assertThat(exception.getMessage(), containsString("Invalid log"));

    }

    @Test
    void testGetSubmissionStatus() {

        // Given
       final SentSubmission sentSubmission = SentSubmission.builder()
               .caseId(UUID.randomUUID())
               .submissionId(UUID.randomUUID())
               .destinationId(UUID.randomUUID())
               .authenticationTags(new AuthenticationTags())
               .build();

        final EventStatus eventStatus = new EventStatus(Event.ACCEPT.getState(), Collections.emptyList());

        when(this.senderMock.getLastedEvent(any(), any(), any(), any())).thenReturn(eventStatus);

        // When
        final EventStatus statusForSubmission = this.underTest.getStatusForSubmission(sentSubmission);

        // Then
        assertThat(statusForSubmission.getStatus(), is(Event.ACCEPT.getState()));

    }

    @Test
    void testGetSubmissionStatusWithProblems() {

        // Given
        final SentSubmission sentSubmission = SentSubmission.builder()
                .caseId(UUID.randomUUID())
                .submissionId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .authenticationTags(new AuthenticationTags())
                .build();

        final EventStatus eventStatus = new EventStatus(Event.REJECT.getState(), List.of(new DataEncryptionIssue(), new DataSchemaViolation()));

        when(this.senderMock.getLastedEvent(any(), any(), any(), any())).thenReturn(eventStatus);

        // When
        final EventStatus statusForSubmission = this.underTest.getStatusForSubmission(sentSubmission);

        // Then
        assertThat(statusForSubmission.getStatus(), is(Event.REJECT.getState()));
        assertThat(statusForSubmission.getProblems(), is(List.of(new DataEncryptionIssue(), new DataSchemaViolation())));

    }

    private UUID setupTestMocks() throws JOSEException {
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(this.senderMock.getDestination(any())).thenReturn(destination);
        when(this.senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(this.senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(this.senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(this.senderMock.sendSubmission(any())).thenReturn(new Submission());

        return destinationId;
    }

    private SubmissionForPickup getAnnouncedSubmission(final UUID destinationId) {
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://json-schema.fitco.de"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("1");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));

        final var announcedSubmission = new SubmissionForPickup();
        announcedSubmission.setSubmissionId(UUID.randomUUID());
        announcedSubmission.setDestinationId(destinationId);
        return announcedSubmission;
    }

    private Destination getDestination(final UUID destinationId) {
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://mimetype.test.de"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private Destination getDestination(final UUID destinationId, final SubmissionSchema schema) {
        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}